@@ -6,6 +6,36 @@ import operator
 from datetime import timedelta, date
 
 
+
+def best_two_distribute_weights_risk(best_two_performers, historic_volatilities, capped_weights, budget_risks_best_performer, target_vol):
+    capped_weights = capped_weights.copy()
+    filtered_capped_weights = {}
+    filtered_historic_volatilities = {}
+    for key, value in capped_weights.items():
+        if key in best_two_performers:
+            filtered_capped_weights[key] = value
+            filtered_historic_volatilities[key] = historic_volatilities[f'previous_rolling_{key}']
+
+    allocated_compo = {}
+    while(len(filtered_capped_weights) > 0):
+        #### we compute the allocation for the biggest component
+        biggest_alloc_constituent = max(filtered_capped_weights, key=filtered_capped_weights.get)
+        biggest_alloc_weight = filtered_capped_weights[biggest_alloc_constituent]
+        if len(filtered_capped_weights) == 1:
+            biggest_alloc_risk_budget = 1.
+        else:
+            biggest_alloc_risk_budget = budget_risks_best_performer
+        histo_vol_biggest_alloc_constituent = filtered_historic_volatilities[biggest_alloc_constituent]
+        biggest_alloc_final_weight = min(biggest_alloc_weight ,(target_vol * biggest_alloc_risk_budget)/ histo_vol_biggest_alloc_constituent)
+        #target_vol = target_vol - target_vol * biggest_alloc_risk_budget
+        target_vol = target_vol - biggest_alloc_final_weight*histo_vol_biggest_alloc_constituent
+        filtered_capped_weights.pop(biggest_alloc_constituent, None)
+        allocated_compo[biggest_alloc_constituent] = biggest_alloc_final_weight
+
+    cash_value = 1-sum(allocated_compo.values())
+    allocated_compo['cash'] = cash_value
+    return allocated_compo
+
 def distribute_weights_risk(historic_volatilities, capped_weights, risk_budget_allocations, target_vol):
     capped_weights = capped_weights.copy()
     allocated_compo = {}
@@ -29,8 +59,118 @@ def distribute_weights_risk(historic_volatilities, capped_weights, risk_budget_a
 def past_volatility(x):
     return np.sqrt(252)*np.std(x)
 
+def follow_dic_order(dict_to_order,dict_to_follow):
+    ordered_dic = {}
+    for key in dict_to_follow.keys():
+        ordered_dic[key] = dict_to_order[key]
+    return ordered_dic
+
+def best_two_indice_rebalancing(price_data=None, target_vol = None, rebalancing_method=None, best_two_window = 252, volatility_window = 252, inception_date = None, capped_weights = None, budget_risks_best_performer = None, initial_value = 100., last= False):
+    assert target_vol is not None
+    assert inception_date >= min(price_data.index) + timedelta(days=volatility_window)
+    assert inception_date >= min(price_data.index) + timedelta(days=best_two_window)
+
+    capped_weights = dict(sorted(capped_weights.items(), key=lambda x: (-x[1], x[0])))
+
+    constituents = list(price_data.columns)
+    constituents.sort()
+    return_underlyings=[]
+    rolling_vol_underlyings=[]
+    previous_rolling_vol_underlyings=[]
+    past_perf_underlyings =[]
+
+    for me_consti in constituents:
+        return_underlying = f'returns_{me_consti}'
+        return_underlyings.append(return_underlying)
+        price_data[return_underlying] = price_data[me_consti].pct_change()
+
+        past_perf_underlying = f'past_perf_{me_consti}'
+        past_perf_underlyings.append(past_perf_underlying)
+        price_data[past_perf_underlying] = price_data[return_underlying].rolling(best_two_window).mean()
+
+        rolling_vol_underlying = f'rolling_{me_consti}'
+        rolling_vol_underlyings.append(rolling_vol_underlying)
+        price_data[rolling_vol_underlying] = price_data[return_underlying].rolling(volatility_window).apply(past_volatility)
+
+        previous_rolling_vol_underlying = f'previous_rolling_{me_consti}'
+        previous_rolling_vol_underlyings.append(previous_rolling_vol_underlying)
+        price_data[previous_rolling_vol_underlying] = price_data[rolling_vol_underlying].shift(1)
+
+        vol_factor_underlying = f'vol_factor_{me_consti}'
+        price_data[vol_factor_underlying] = target_vol / price_data[previous_rolling_vol_underlying]
+
+    print(f'date range before filtering {min(price_data.index)} {max(price_data.index)}')
+    price_data = price_data.loc[price_data.index >= inception_date, :]
+    print(f'date range after filtering {min(price_data.index)} {max(price_data.index)}')
+    assert inception_date in price_data.index
+
+
+    date_utility.add_rebalancing_datepart(price_data, 'Date', rebalancing_method = rebalancing_method, last=last)
+
+
+    weights_list = []
+
+    counter = 0
+    previous_value = None
+    for date_index, row in price_data.iterrows():
+        must_rebalance = row['is_rebalancing']
+        current_returns = price_data.loc[date_index, return_underlyings].to_dict()
+        if must_rebalance or counter == 0:
+            past_perf = price_data.loc[date_index, past_perf_underlyings]
+            past_perf = past_perf.sort_values(ascending=False)
+            best_two_performers = list(past_perf.index[:2])
+            best_two_performers = [x.replace('past_perf_', '') for x in best_two_performers]
+
+            previous_histo_volatilities = price_data.loc[date_index, previous_rolling_vol_underlyings]
+            current_weights = best_two_distribute_weights_risk(best_two_performers,
+                                                               previous_histo_volatilities.to_dict(), capped_weights,
+                                                               budget_risks_best_performer, target_vol)
+            if previous_value is None:
+                value = initial_value
+            else:
+                value = previous_value * (global_drift_weight(previous_weights, current_returns) / sum(previous_weights.values()))
+        else :
+            value = previous_value * (global_drift_weight(previous_weights,current_returns)/sum(previous_weights.values()))
+            current_weights = drift_weight(previous_weights,current_returns)
+        counter=counter+1
+
+        previous_value = value
+        previous_weights = current_weights.copy()
+        iteration = current_weights.copy()
+        iteration.update({'Date' : date_index, 'indice' : value})
+        weights_list.append(iteration)
+
+    weights_df = pd.DataFrame(weights_list)
+    weights_df.index = pd.to_datetime(weights_df['Date'])
+    date_utility.add_rebalancing_datepart(weights_df, 'Date', rebalancing_method = rebalancing_method, last = last)
+
+    return weights_df
+# ################# numba version
+#     starting_date = inception_date - timedelta(days=best_two_window)
+#     price_data = price_data.loc[price_data.index >= starting_date, :]
+#
+#     for t in range(best_two_window, price_data.shape[0]):
+#         past_price_df = price_data.iloc[t-best_two_window:t,:].copy()
+#         past_perf = past_price_df[return_underlyings].mean(axis = 0)
+#         past_perf = past_perf.sort_values(ascending=False)
+#         best_two_performers = list(past_perf.index[:2])
+#         best_two_performers = [ x.replace('returns_','') for x in best_two_performers]
+#         date_index = price_data.index[t]
+#         previous_histo_volatilities = price_data.loc[date_index, previous_rolling_vol_underlyings]
+#         current_weights = best_two_distribute_weights_risk(best_two_performers, previous_histo_volatilities.to_dict(), capped_weights,
+#                                                            risk_budget_allocations, target_vol)
+#
+
+
+
+
+
+
 def vol_target_indice_rebalancing(weight_data=None, vol_target = None, volatility_window = 252, vol_target_inception_date = None, initial_value = 100.):
     assert vol_target is not None
+
+    constituents = [col for col in weight_data.columns if col not in ['is_rebalancing', 'indice']]
+    constituents.sort()
     weight_data['indice_returns'] = weight_data['indice'].pct_change()
     weight_data['rolling_vol_indice'] = weight_data['indice_returns'].rolling(volatility_window).apply(past_volatility)
     is_past_volatility_not_available_at_start =  np.isnan(weight_data.loc[vol_target_inception_date,'rolling_vol_indice'])
@@ -43,7 +183,6 @@ def vol_target_indice_rebalancing(weight_data=None, vol_target = None, volatilit
     target_vol_values_list = []
     for date_index, row in weight_data.iterrows():
         if date_index == vol_target_inception_date:
-            assert row['is_rebalancing']
             volatility_factor = vol_target / weight_data.loc[vol_target_inception_date,'rolling_vol_indice']
             previous_value = initial_value
             target_vol_values_list.append({'Date': vol_target_inception_date, 'indice': initial_value, 'volatility_factor':np.nan})
@@ -60,7 +199,16 @@ def vol_target_indice_rebalancing(weight_data=None, vol_target = None, volatilit
     target_vol_values_df = pd.DataFrame(target_vol_values_list)
     target_vol_values_df.index = pd.to_datetime(target_vol_values_df['Date'])
     target_vol_values_df.drop('Date', axis=1, inplace=True)
-    return target_vol_values_df
+
+    accumulated_lost_weight = np.zeros(weight_data.shape[0])
+    for me_consti in constituents:
+        if me_consti != 'cash':
+            accumulated_lost_weight = accumulated_lost_weight + weight_data[me_consti].values
+            weight_data[me_consti] = weight_data[me_consti]*target_vol_values_df['volatility_factor']
+
+    weight_data['cash'] = weight_data['cash']+(1-target_vol_values_df['volatility_factor'])*accumulated_lost_weight
+    weight_data = weight_data.drop(columns=['indice_returns', 'rolling_vol_indice'])
+    return target_vol_values_df, weight_data
 
 def drift_weight(previous_weights, current_returns):
     drifted_weights = previous_weights.copy()
@@ -137,7 +285,6 @@ def vol_saturation_indice_rebalancing(price_data= None, inception_date=None, ris
 
     weights_list = []
 
-    current_weights = {}
 
     previous_value = None
     for date_index, row in price_data.iterrows():
@@ -189,7 +336,7 @@ def drop_budget_risk_to_key(current_weights):
         translated_weights[key.replace('budget_risk_', '')] = val
     return translated_weights
 
-def indice_rebalancing(price_data= None, inception_date=None, target_weights=None, initial_value= 100., rebalancing_method = None , early_rebalancing = False):
+def indice_rebalancing(price_data= None, inception_date=None, target_weights=None, initial_value= 100., rebalancing_method = None , early_rebalancing = False, last = True):
     if target_weights is None:
         target_weights = {}
     constituents = target_weights.keys()
@@ -207,7 +354,7 @@ def indice_rebalancing(price_data= None, inception_date=None, target_weights=Non
     price_data = price_data.loc[price_data.index >= inception_date, :]
     print(f'date range after filtering {min(price_data.index)} {max(price_data.index)}')
     assert inception_date in price_data.index
-    date_utility.add_rebalancing_datepart(price_data, 'Date', rebalancing_method = rebalancing_method)
+    date_utility.add_rebalancing_datepart(price_data, 'Date', rebalancing_method = rebalancing_method,last = last)
     assert price_data.shape[0] > 0
     initial_weights = target_weights.copy()
     initial_weights.update({
@@ -255,7 +402,7 @@ def indice_rebalancing(price_data= None, inception_date=None, target_weights=Non
 
     weights_df = pd.DataFrame(weights_list)
     weights_df.index = pd.to_datetime(weights_df['Date'])
-    date_utility.add_rebalancing_datepart(weights_df, 'Date', rebalancing_method = rebalancing_method)
+    date_utility.add_rebalancing_datepart(weights_df, 'Date', rebalancing_method = rebalancing_method, last = last)
 
     for me_constituent in constituents:
         weights_df[me_constituent] = weights_df[me_constituent] / weights_df['indice']
